/**
 * Copyright (C) 2007-2011 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <glib/gstdio.h>

#include <libsoup/soup.h>

#include "main.h"
#include "wk-pl-parser.h"
#include "wk-slideshow-item.h"
#include "wk-settings.h"
#include "wk-daemon.h"
#include "wk-download.h"
#include "wk-browser.h"
#include "wk-events.h"
#include "text-utils.h"
#include "enums.h"
#include "utils.h"

static void _daemon_select_playlist_loop (WkDaemon *d);

static const gchar *todownloadprefixes[] = {
    "http://",
    "ftp://",
    "https://",
    NULL
};

static gboolean
to_download_prefix (const gchar *uri)
{
    gboolean ret = FALSE;
    const gchar **p;

    for (p=todownloadprefixes;*p;p++)
    {
        if (g_str_has_prefix (uri, *p))
        {
            ret = TRUE;
            break;
        }
    }

    return ret;
}

static gboolean
_daemon_browser_show_uri (WkDaemon *d, const gchar *uri)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    DEBUG ("showing item: %s", uri);

    return wk_browser_show (browser, WK_SLIDESHOW_ITEM_TYPE_WEB, uri);
}

static gboolean
_daemon_browser_show (WkDaemon *d, WkSlideshowItem *item)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    DEBUG ("showing item: %s", item->uri);

    return wk_browser_show (browser, item->type, item->uri);
}

static gboolean
_daemon_show_next_item (gpointer data)
{
    WkSlideshowItem *item;
    WkDaemon *d = (WkDaemon *) data;

    DEBUG ("start");

    if (d->playlist && ((d->current = g_slist_next (d->current)) == NULL))
    {
        DEBUG ("end of playlist loop");
        _daemon_select_playlist_loop (d);
    }

    if (! d->current)
    {
        if (d->next_playlist && d->do_pl_switch)
        {
            DEBUG ("switching playlist");
            g_atomic_int_set (&d->do_pl_switch, 0);
            if (d->playlist)
                wk_pl_free_playlist (d->playlist);
            d->playlist = d->next_playlist;
            d->next_playlist = NULL;
        }
        d->current = d->playlist;
    }

    item = (WkSlideshowItem *) d->current->data;

    /* if href attribute is true, this is only data used by other items. We
       don't show it */
    if (item->href)
    {
        DEBUG ("not showing href resource: %s", item->uri);
        _daemon_show_next_item (data);
        return FALSE;
    }

    /* Deal with idle. */
    if (item->type == WK_SLIDESHOW_ITEM_TYPE_IDLE)
    {
        if (d->status != WK_DAEMON_STATUS_IDLE)
        {
            DEBUG ("going to idle");
            g_spawn_command_line_async ("xset dpms force off", NULL);
            d->status = WK_DAEMON_STATUS_IDLE;
        }
        d->timeout_id = g_timeout_add_seconds (item->timeout,
                                               _daemon_show_next_item,
                                               d);
        return FALSE;
    }

    if (d->status == WK_DAEMON_STATUS_IDLE)
    {
        DEBUG ("exiting idle");
        d->status = WK_DAEMON_STATUS_SLIDESHOW;
        g_spawn_command_line_async ("xset dpms force on", NULL);
    }

    if (!_daemon_browser_show (d, item))
    {
        _daemon_show_next_item (d);
        return FALSE;
    }

    if ((item->type != WK_SLIDESHOW_ITEM_TYPE_VIDEO) && !is_video_uri (item->uri))
    {
        d->timeout_id = g_timeout_add_seconds (item->timeout,
                                               _daemon_show_next_item,
                                               d);
    }

    DEBUG ("end");

    return FALSE;
}

void
_daemon_browser_stop_session (WkDaemon *d)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    DEBUG ("session stop");

    wk_browser_clear_session (browser, TRUE);
    d->status = WK_DAEMON_STATUS_INVALID;
}

void
_daemon_browser_start_session (WkDaemon *d)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    if (d->status == WK_DAEMON_STATUS_BROWSER)
        return;

    if (d->status == WK_DAEMON_STATUS_IDLE)
        g_spawn_command_line_async ("xset dpms force on", NULL);

    DEBUG ("session start");

    if (d->settings->slideshow_hide_pointer)
        stop_unclutter ();

    wk_browser_start_session (browser);
    d->status = WK_DAEMON_STATUS_BROWSER;
}

void
_daemon_handle_post_session (WkDaemon *d)
{
    switch (d->settings->after_web_session)
    {
        case AFTER_WEB_SESSION_SLIDESHOW:
            wk_daemon_start_slideshow (d);
            break;
        case AFTER_WEB_SESSION_SESSION:
            if (d->settings->with_timer)
                _daemon_browser_show_uri (d, d->settings->default_url);
            else
                _daemon_browser_start_session (d);
            break;
        default:
            _daemon_browser_show_uri (d, d->settings->default_url);
            break;
    }
}

gboolean
_daemon_post_session_timeout (gpointer data)
{
    WkDaemon *d = (WkDaemon *) data;
    d->status = WK_DAEMON_STATUS_INVALID;

    _daemon_handle_post_session ((WkDaemon *) data);

    return FALSE;
}

static gpointer
_daemon_get_cache_thread (gpointer data)
{
    WkDaemon *d = (WkDaemon *) data;
    GSList *p;
    gboolean success;
    gchar *tmp;
    gchar *target;
    gchar *etag_file;

    DEBUG ("start");

    for (p = d->next_playlist; p; p = p->next)
    {
        WkSlideshowItem *item = (WkSlideshowItem *) p->data;

        if (item->type == WK_SLIDESHOW_ITEM_TYPE_IDLE)
        {
            DEBUG ("First item is idle");
            break;
        }

        if (item->type == WK_SLIDESHOW_ITEM_TYPE_WEB)
            continue;

        if (!to_download_prefix (item->uri))
            continue;

        tmp = g_path_get_basename (item->uri);
        target = g_strconcat (d->data_path, tmp, NULL);
        etag_file = g_strconcat (d->etag_path, tmp, NULL);
        g_free (tmp);

        if (!g_file_test (target, G_FILE_TEST_EXISTS))
        {
            success = wk_download_perform (item->uri, target, etag_file);
            DEBUG ("downloaded: %s => %s (%d)", item->uri, target, success);
        }
        else if (wk_download_is_needed (item->uri, target, etag_file))
        {
            success = wk_download_perform (item->uri, target, etag_file);
            DEBUG ("re-downloaded: %s => %s (%d)", item->uri, target, success);
        }
        else
            success = TRUE;

        if (success)
        {
            g_free (item->uri);
            if (item->type == WK_SLIDESHOW_ITEM_TYPE_TEXT)
            {
                /* update the text with local data */
                text_fill_cache (target, d->data_path);
                item->uri = g_strconcat ("file://", target, NULL);
                g_free (target);
            }
            else
            {
                item->uri = target;
            }
        } else {
            item->type = WK_SLIDESHOW_ITEM_TYPE_WEB;
        }
    }

    g_atomic_int_set (&d->do_pl_switch, 1);

    DEBUG ("end");

    return NULL;
}

static void
_daemon_get_cache (WkDaemon *d, gboolean threaded)
{
    if (threaded)
        g_thread_create (_daemon_get_cache_thread, d, FALSE, NULL);
    else
        _daemon_get_cache_thread (d);
}

static void
_daemon_get_playlist_data (WkDaemon *d)
{
    DEBUG ("start");

    if (d->playlist_data)
        g_free (d->playlist_data);

    /* start loading the slideshow infos if any */
    DEBUG ("slideshow_file: %s", d->settings->slideshow_file);
    if (d->settings->slideshow_file[0] == '/')
        g_file_get_contents (d->settings->slideshow_file,
                             &d->playlist_data, NULL, NULL);
    else if (strncmp (d->settings->slideshow_file, "file://", 7) == 0)
        g_file_get_contents (d->settings->slideshow_file + 7,
                             &d->playlist_data, NULL, NULL);
    else
        d->playlist_data = get_remote_data (d->settings->slideshow_file);

    if (!d->playlist_data)
    {
        DEBUG ("getting XML data from cache");
        if (!g_file_get_contents (d->playlist_file, &d->playlist_data, NULL, NULL))
            d->playlist_data = NULL;
    }
}

static void
_daemon_select_playlist_loop (WkDaemon *d)
{
    DEBUG ("start");

    if (d->playlist_data)
    {
        DEBUG ("reparsing playlist");
        d->next_playlist = wk_pl_get_playlist (d->playlist_data);
        if (d->next_playlist)
        {
            DEBUG ("new playlist will be used");
            /* store the playlist for next use */
            g_file_set_contents (d->playlist_file, d->playlist_data, -1, NULL);

            WkSlideshowItem *item = (WkSlideshowItem *) d->next_playlist->data;
            if (item->type != WK_SLIDESHOW_ITEM_TYPE_IDLE)
            {
                //_daemon_get_cache (d, TRUE);
                _daemon_get_cache (d, FALSE);
            }
            else
                g_atomic_int_set (&d->do_pl_switch, 1);

        }
    }
}

static gboolean
_on_playlist_timeout (WkDaemon *d)
{
    _daemon_get_playlist_data (d);

    return TRUE;
}

WkDaemon *
wk_daemon_get ()
{
    static WkDaemon *d = NULL;
    const gchar *home;

    if (d)
        return d;

    d = g_new (WkDaemon, 1);
    d->status = WK_DAEMON_STATUS_INVALID;
    d->playlist = NULL;
    d->next_playlist = NULL;
    d->do_pl_switch = 0;
    d->current = NULL;
    d->settings = GET_SETTINGS ();

    home = getenv ("HOME");
    d->data_path = g_strdup_printf ("%s/.linutop-kiosk/data/", home);
    g_mkdir_with_parents (d->data_path, 0755);
    d->etag_path = g_strdup_printf ("%s/.linutop-kiosk/etag/", home);
    g_mkdir_with_parents (d->etag_path, 0755);
    d->playlist_file = g_strdup_printf ("%s/.linutop-kiosk/playlist", home);

    d->playlist_data = NULL;

    return d;
}

void
wk_daemon_start (WkDaemon *d)
{
    if (d->status != WK_DAEMON_STATUS_INVALID)
        return;

    if (d->settings->mode == MODE_KIOSK)
    {
        if (d->settings->prestart_file &&
               (g_access (d->settings->prestart_file, X_OK) == 0))
        {
            gint status;

            if (g_spawn_command_line_sync (d->settings->prestart_file,
                                       NULL, NULL, &status, NULL)
                && WIFEXITED (status)
                && (WEXITSTATUS (status) == 0))
            {
                _daemon_browser_start_session (d);
            }
        }
        else if (d->settings->start_with_session)
            _daemon_browser_start_session (d);
    }

    /* in display or kiosk+display mode only */
    if ((d->settings->mode != MODE_KIOSK)
        || (d->settings->after_web_session == AFTER_WEB_SESSION_SLIDESHOW))
    {
        _daemon_get_playlist_data (d);
        _daemon_select_playlist_loop (d);
    }

    if (d->settings->slideshow_refresh_after_minutes)
    {
        DEBUG ("refreshing playlist every %d minutes",
               d->settings->slideshow_refresh_after_minutes);
        g_timeout_add_seconds (d->settings->slideshow_refresh_after_minutes*60,
                               (GSourceFunc) _on_playlist_timeout, d);
    }

    /* stop here if we've started something */
    if (d->status != WK_DAEMON_STATUS_INVALID)
        return;

    if (d->next_playlist)
    {
        DEBUG ("starting slideshow");
        wk_daemon_start_slideshow (d);
    }
    else
    {
        DEBUG ("displaying default URL");
        _daemon_browser_show_uri (d, d->settings->default_url);
    }
}

void
wk_daemon_start_slideshow (WkDaemon *d)
{
    DEBUG ("start");

    if (d->status == WK_DAEMON_STATUS_SLIDESHOW)
        return;
    if (d->status == WK_DAEMON_STATUS_IDLE)
        g_spawn_command_line_async ("xset dpms force on", NULL);
    if (d->status == WK_DAEMON_STATUS_BROWSER)
        _daemon_browser_stop_session (d);

    if (d->settings->slideshow_hide_pointer)
        start_unclutter ();

    /* deal with the playlist not being ready */
    if (d->next_playlist && !d->playlist)
    {
        DEBUG ("waiting for playlist setup");
        while (g_atomic_int_get (&d->do_pl_switch) != 1)
            g_usleep (100000);
        DEBUG ("done");
    }

    d->status = WK_DAEMON_STATUS_SLIDESHOW;
    d->current = NULL;

    wk_event_set_event_filter ();

    _daemon_show_next_item (d);

    DEBUG ("end");
}

void
wk_daemon_handle_start_session (WkDaemon *d)
{
    GSource *source;

    DEBUG ("start");

    if (d->settings->mode != MODE_KIOSK)
        return;

    if (d->status == WK_DAEMON_STATUS_BROWSER)
        return;

    if (d->status == WK_DAEMON_STATUS_IDLE)
        g_spawn_command_line_async ("xset dpms force on", NULL);

    d->status = WK_DAEMON_STATUS_INVALID;

    if (d->timeout_id
        && (source = g_main_context_find_source_by_id (
                    NULL, d->timeout_id)) != NULL)
    {
        g_source_destroy (source);
    }

    _daemon_browser_start_session (d);

    DEBUG ("end");
}

void
wk_daemon_handle_stop_session (WkDaemon *d)
{
    if (d->status != WK_DAEMON_STATUS_BROWSER)
        return;

    wk_browser_clear_session (WK_BROWSER (wk_browser_get (NULL)), TRUE);
    d->status = WK_DAEMON_STATUS_INVALID;

    if (d->settings->post_session_uri)
    {
        _daemon_browser_show_uri (d, d->settings->post_session_uri);
        g_timeout_add_seconds (d->settings->post_session_uri_time,
                               _daemon_post_session_timeout,
                               d);
    }
    else
        _daemon_handle_post_session (d);
}

void
wk_daemon_handle_video_end (WkDaemon *d)
{
    if (d->status != WK_DAEMON_STATUS_SLIDESHOW)
        return;

    _daemon_show_next_item (d);
}
